///////////////////////////////////////////////////////////////////////////
// Copyright (C) 2008  Whit Armstrong                                    //
//                                                                       //
// This program is free software: you can redistribute it and/or modify  //
// it under the terms of the GNU General Public License as published by  //
// the Free Software Foundation, either version 3 of the License, or     //
// (at your option) any later version.                                   //
//                                                                       //
// This program is distributed in the hope that it will be useful,       //
// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         //
// GNU General Public License for more details.                          //
//                                                                       //
// You should have received a copy of the GNU General Public License     //
// along with this program.  If not, see <http://www.gnu.org/licenses/>. //
///////////////////////////////////////////////////////////////////////////

#ifndef VECTOR_SUMMARY_HPP
#define VECTOR_SUMMARY_HPP

// taking 1 vector an argument
#include <tslib/vector.summary/max.hpp>
#include <tslib/vector.summary/min.hpp>
#include <tslib/vector.summary/mean.hpp>
#include <tslib/vector.summary/sum.hpp>
#include <tslib/vector.summary/prod.hpp>
#include <tslib/vector.summary/stdev.hpp>
#include <tslib/vector.summary/rank.hpp>
#include <tslib/vector.summary/open.hpp>
#include <tslib/vector.summary/close.hpp>

// technical analysis
#include <tslib/vector.summary/rsi.hpp>

// taking 2 vectors as arguments
#include <tslib/vector.summary/cov.hpp>
#include <tslib/vector.summary/cor.hpp>

#endif // VECTOR_SUMMARY_HPP
